"""
Common utilities needed that make it possible to convert actions (e.g. choosing
threes on the scorecard) to categories (e.g. full house on the scorecard) and vice versa.
Similar maps are also provided for mapping dice rolling lists (e.g. reroll dice 2, 3 and
4 but keep dice 1 and 5) to and from actions and categories to scoring functions.
"""
from typing import Dict, List, Tuple

from pyhtzee.classes import Category, Rule

# Mapping from action id to permutations of rerolling of dice. Each unique combination
# of dice rolls is given a unique id, resulting in 32 unique constellations. However,
# keeping all dice is left out, as the player should then choose a category from the
# scorecard, hence there are 31 unique rerolling patterns. An offset constant is
# provided, as category actions are located after dice rolling actions.
CATEGORY_ACTION_OFFSET = 31
action_to_dice_roll_map: Dict[int, Tuple[bool, bool, bool, bool, bool]] = {}
dice_roll_to_action_map: Dict[Tuple[bool, bool, bool, bool, bool], int] = {}
for d1 in [1, 0]:
    for d2 in [1, 0]:
        for d3 in [1, 0]:
            for d4 in [1, 0]:
                for d5 in [1, 0]:
                    # make rolling all dice the first action, i.e. zero
                    key = 31 - (d5 * 2**0 + d4 * 2**1 + d3 * 2**2 + d2 * 2**3 + d1 * 2**4)
                    value = bool(d1), bool(d2), bool(d3), bool(d4), bool(d5)
                    # not rolling any dice is not a valid action
                    if key < 31:
                        action_to_dice_roll_map[key] = value
                        dice_roll_to_action_map[value] = key


# Mapping from action id to category and vice versa.
action_to_category_map: Dict[int, Category] = {}
category_to_action_map: Dict[Category, int] = {}
for i, category in enumerate(Category):
    if category in (Category.UPPER_SECTION_BONUS, Category.YAHTZEE_BONUS):
        continue
    action_to_category_map[i + CATEGORY_ACTION_OFFSET] = category
    category_to_action_map[category] = i + CATEGORY_ACTION_OFFSET

# List of actionable categories e.g. for determining valid actions
actionable_categories: List[Category] = []
for category in Category:
    if category not in (Category.UPPER_SECTION_BONUS, Category.YAHTZEE_BONUS):
        actionable_categories.append(category)


def is_upper_section_category(category: Category) -> bool:
    return True if int(category) < 6 else False


def is_joker_category(category: Category) -> bool:
    return True if category in [
        Category.FULL_HOUSE,
        Category.SMALL_STRAIGHT,
        Category.LARGE_STRAIGHT
    ] else False


def is_category_supported_by_rule(category: Category, rule: Rule) -> bool:
    if rule != Rule.YATZY and (category == Category.ONE_PAIR
                               or category == Category.TWO_PAIRS):
        return False
    return True



"""
Scoring functions for lists with five dice. All functions (except upper sectino bonus)
are given a a list of ints representing the number of dots on the face. If the list of
dice is not valid for the scoring function, a zero is retured, indicating that the user
has stricken that score box off the scorecard.
"""
from collections import Counter
from typing import Dict, List, Set

from pyhtzee.classes import Category, Rule

CONSTANT_SCORES: Dict[Category, int] = {
    Category.YAHTZEE: 50,
}

CONSTANT_SCORES_YAHTZEE: Dict[Category, int] = {
    Category.FULL_HOUSE: 25,
    Category.SMALL_STRAIGHT: 30,
    Category.LARGE_STRAIGHT: 40,
    Category.YAHTZEE_BONUS: 100,
    Category.UPPER_SECTION_BONUS: 35,
}

CONSTANT_SCORES_YATZY: Dict[Category, int] = {
    Category.UPPER_SECTION_BONUS: 50,
}


def score_upper_section(dice: List[int], face: int) -> int:
    return sum(die if die == face else 0 for die in dice)


def score_x_of_a_kind(dice: List[int], min_same_faces: int, rule: Rule) -> int:
    if rule == Rule.YATZY:
        return score_x_of_a_kind_yatzy(dice, min_same_faces)
    return score_x_of_a_kind_yahtzee(dice, min_same_faces)


def score_x_of_a_kind_yahtzee(dice: List[int], min_same_faces: int) -> int:
    """Return sum of dice if there are a minimum of equal min_same_faces dice, otherwise
    return zero. Only works for 3 or more min_same_faces.
    """
    for die, count in Counter(dice).most_common(1):
        if count >= min_same_faces:
            return sum(dice)
    return 0


def score_x_of_a_kind_yatzy(dice: List[int], min_same_faces: int) -> int:
    """Similar to yahtzee, but only return the sum of the dice that satisfy min_same_faces
    """
    for die, count in Counter(dice).most_common(1):
        if count >= min_same_faces:
            return die * min_same_faces
    return 0


def score_full_house(dice: List[int], rule: Rule) -> int:
    global CONSTANT_SCORES_YAHTZEE
    counter = Counter(dice)
    if len(counter.keys()) == 2 and min(counter.values()) == 2:
        if rule == Rule.YATZY:
            return sum(dice)
        return CONSTANT_SCORES_YAHTZEE[Category.FULL_HOUSE]
    return 0


def _are_two_sets_equal(a: Set, b: Set) -> bool:
    return a.intersection(b) == a


def score_one_pair(dice: List[int]) -> int:
    pairs: Set[int] = set()
    for die, count in Counter(dice).most_common():
        if count >= 2:
            pairs.add(die)
    if pairs:
        sorted_pairs = sorted(pairs, reverse=True)
        return sorted_pairs[0] * 2
    return 0


def score_two_pairs(dice: List[int]) -> int:
    pairs: Set[int] = set()
    for die, count in Counter(dice).most_common():
        if count >= 2:
            pairs.add(die)
    if len(pairs) >= 2:
        sorted_pairs = sorted(pairs, reverse=True)
        return sorted_pairs[0] * 2 + sorted_pairs[1] * 2
    return 0


def score_small_straight(dice: List[int], rule: Rule) -> int:
    if rule == Rule.YATZY:
        return score_small_straight_yatzy(dice)
    return score_small_straight_yahztee(dice)


def score_small_straight_yahztee(dice: List[int]) -> int:
    """
    Small straight scoring according to regular yahtzee rules
    """
    global CONSTANT_SCORES_YAHTZEE
    dice_set = set(dice)
    if _are_two_sets_equal({1, 2, 3, 4}, dice_set) or \
            _are_two_sets_equal({2, 3, 4, 5}, dice_set) or \
            _are_two_sets_equal({3, 4, 5, 6}, dice_set):
        return CONSTANT_SCORES_YAHTZEE[Category.SMALL_STRAIGHT]
    return 0


def score_small_straight_yatzy(dice: List[int]) -> int:
    """
    Small straight scoring according to yatzy rules
    """
    dice_set = set(dice)
    if _are_two_sets_equal({1, 2, 3, 4, 5}, dice_set):
        return sum(dice)
    return 0


def score_large_straight(dice: List[int], rule: Rule) -> int:
    if rule == Rule.YATZY:
        return score_large_straight_yatzy(dice)
    else:
        return score_large_straight_yahtzee(dice)


def score_large_straight_yahtzee(dice: List[int]) -> int:
    """
    Large straight scoring according to regular yahtzee rules
    """
    global CONSTANT_SCORES_YAHTZEE
    dice_set = set(dice)
    if _are_two_sets_equal({1, 2, 3, 4, 5}, dice_set) or \
            _are_two_sets_equal({2, 3, 4, 5, 6}, dice_set):
        return CONSTANT_SCORES_YAHTZEE[Category.LARGE_STRAIGHT]
    return 0


def score_large_straight_yatzy(dice: List[int]) -> int:
    """
    Large straight scoring according to yatzy rules
    """
    dice_set = set(dice)
    if _are_two_sets_equal({2, 3, 4, 5, 6}, dice_set):
        return sum(dice)
    return 0


def score_yahtzee(dice: List[int]) -> int:
    global CONSTANT_SCORES_YAHTZEE, CONSTANT_SCORES_YATZY
    if len(set(dice)) == 1:
        return CONSTANT_SCORES[Category.YAHTZEE]
    return 0


def score_chance(dice: List[int]) -> int:
    return sum(dice)


def score_upper_section_bonus(upper_section_score: int, rule: Rule) -> int:
    global CONSTANT_SCORES_YAHTZEE, CONSTANT_SCORES_YATZY
    if upper_section_score >= 63:
        if rule == Rule.YATZY:
            return CONSTANT_SCORES_YATZY[Category.UPPER_SECTION_BONUS]
        else:
            return CONSTANT_SCORES_YAHTZEE[Category.UPPER_SECTION_BONUS]
    return 0



import io

from setuptools import find_packages, setup

with io.open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='pyhtzee',
    version='1.2.5',
    description='Yahtzee game engine',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Ville Brofeldt',
    author_email='ville.brofeldt@iki.fi',
    maintainer='Ville Brofeldt',
    maintainer_email='ville.brofeldt@iki.fi',
    url='https://github.com/villebro/pyhtzee',
    license='MIT',
    packages=find_packages(exclude='tests'),
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
    ],
)



"""
Common classes needed for gameplay
"""
from enum import Enum, IntEnum


class PyhtzeeException(Exception):
    pass


class Category(IntEnum):
    ACES = 0
    TWOS = 1
    THREES = 2
    FOURS = 3
    FIVES = 4
    SIXES = 5
    THREE_OF_A_KIND = 6
    FOUR_OF_A_KIND = 7
    FULL_HOUSE = 8
    SMALL_STRAIGHT = 9
    LARGE_STRAIGHT = 10
    YAHTZEE = 11
    CHANCE = 12
    ONE_PAIR = 13
    TWO_PAIRS = 14
    UPPER_SECTION_BONUS = 15
    YAHTZEE_BONUS = 16


class Rule(Enum):
    YAHTZEE = 0
    YAHTZEE_FREE_CHOICE_JOKER = 1
    YATZY = 2
